# CanopyView 2D Plugins - schematic.py
# Description: Schematic plugin.
# Contributors: Nik Molnar
# Created: 2009-06-13
# (c)2009 The Evergreen State College

import math
from PIL import Image
from Plugins import *
from OmniImage import *

#Hard-coded options
kPixelsToMeter        = 10 #Number of pixels in one meter
kCellSize            = 50 #Size of one grid square in pixels
kBaseImageSize        = (300,875) #Should be changed later to take user-input or use the largest tree in the dataset.
kBranchAngle        = .436 #radians - 25 degrees!
kDividingLine       = 1

def init():
   return DrawingPlugin()

class DrawingPlugin:
   "Drawing plugin class."

   def __init__(self):
       "Constructor."

       self.dataset        = None
       self.stemsById        = { }

   def RequiredType(self):
       "Returns 'sql', 'flat', or None depending on required dataset type."

   def RequiredTables(self):
       "Returns a list of required tables, or None if expecting a table-less dataset."

       return ["stem", "branch", "reiteration"]

   def RequiredColumns(self):
       "Returns a list of required columns in the form [(table, column), ...]"

       return [("reiteration", "reiteration_id"), ("reiteration", "stem_id"),
               ("reiteration", "site_id"), ("reiteration","base_ht"), ("reiteration", "top_ht"),
               ("reiteration", "base azi"), ("reiteration", "top_azi"),
               ("reiteration", "basal_diameter"), ("reiteration","base dist"),
               ("reiteration", "top_dist"), ("stem", "site_id"), ("stem", "stem_id"),
               ("stem", "dbh"), ("stem", "height"), ("stem", "branch_data?"),
               ("stem", "live_status"), ("branch", "site_id"), ("branch", "stem_id"),
               ("branch", "branch_id"), ("branch", "live_status"),("branch", "height"),
               ("branch", "diameter"), ("branch", "azimuth"),("branch", "extent"),
               ("branch", "slope"), ("branch", "epicormic")]




   def GetStem(self, dataset, siteId, stemId):
       "Internal method. Get a single tree and associated branches for the given dataset."

       result    = dataset.ExecQuery("SELECT * FROM stem WHERE site_id = \"%s\" AND stem_id = %d" % (siteId, stemId))
       row        = result.NextRowD()

       if not row:
           return None

       stem = Stem()
       if row['stem_id']:            stem.id             = int(row['stem_id'])
       if row['site_id']:            stem.site            = row['site_id']
       if row['plot_id']:            stem.plot             = int(row['plot_id'])
       if row['species_id']:        stem.species        = row['species_id']
       if row['height']:            stem.height            = float(row['height'])
       if row['dbh']:                stem.diameter        = float(row['dbh'])
       if row['live_status']:        stem.alive            = (row['live_status'].lower() == "l")
       if row['branch_data?']:        stem.hasBranchData    = (row['branch_data?'] == "1" or row['branch_data?'] == True)

       result = dataset.ExecQuery("SELECT * FROM branch WHERE site_id = \"%s\" AND stem_id = %d" % (siteId, stemId))
       while 1:
           row = result.NextRowD()

           if not row:
               break

           branch = Branch()
           if row['branch_id']:        branch.id            = int(row['branch_id'])
           if row['stem_id']:            branch.stem            =int(row['stem_id'])
           if row['site_id']:            branch.site            = row['site_id']
           if row['height']:            branch.height        = float(row['height'])
           if row['extent']:            branch.extent        = float(row['extent'])
           if row['azimuth']:            branch.azimuth        = int(row['azimuth'])
           if row['diameter']:            branch.diameter        = float(row['diameter']) / 100
           if row['live_status']:        branch.alive        = (row['live_status'].lower() == "l")
           if row['slope']:            branch.slope        = int(row['slope'])
           if row['epicormic']:        branch.epicormic    = (row['epicormic'] == "epicormic")
           if row['percent_cover']:    branch.percentCover = int(row['percent_cover'])

           #if row['Comes_from']:    branch.comes_from = row['Comes_from']
           branch.comes_from = row.get('comes_from', "M")

           stem.branches.append(branch)

       result = dataset.ExecQuery(
                   "SELECT * FROM reiteration WHERE site_id = \"%s\" AND stem_id = %d" % (siteId, stemId))
       while 1:
           row = result.NextRowD()

           if not row:
               break

           reit = Reiteration()
           if row['reiteration_id']:    reit.id            =int(row['reiteration_id'])
           if row['stem_id']:            reit.stem        = int(row['stem_id'])
           if row['site_id']:            reit.site        = row['site_id']
           if row['base azi']:            reit.azimuth    =int(row['base azi'])
           if row['base_ht']:            reit.base        =float(row['base_ht'])
           if row['top_ht']:            reit.top        = float(row['top_ht'])
           if row['basal_diameter']:    reit.diameter    =float(row['basal_diameter']) / 100
           if row['base dist']:        reit.distance    =float(row['base dist'])
           if row['reit_name']:           reit.name       = row['reit_name']

           if row['top_azi'] and not row['base azi']:reit.azimuth    = int(row['top_azi'])
           if row['top_dist'] and not row['base dist']:reit.distance    = float(row['top_dist'])

           stem.reiterations.append(reit)

       return stem

   def Info(self):
       "Plugin information."

       info = {
           'name'            : "Reiteration Drawing",
           'description'    : "Generates 2D images representing pen and ink drawings of trees.",
           'unit_label'    : "Stem",
           'views'            : 1
       }

       return info

   def ValidData(self, dataset):
       '''Returns True if the provided dataset contains the data
necessary to generate this visualization.
          False otherwise.'''

       #Check Tables
       if not dataset.HasTable("stem", True):
           return False

       if not dataset.HasTable("branch", True):
           return False

       if not dataset.HasTable("reiteration", True):
           return False

       #Reiteration Table
       if not dataset.HasColumn("reiteration", "reiteration_id", True):
           return False

       if not dataset.HasColumn("reiteration", "stem_id", True):
           return False

       if not dataset.HasColumn("reiteration", "site_id", True):
           return False

       if not dataset.HasColumn("reiteration", "base_ht", True):
           return False

       if not dataset.HasColumn("reiteration", "top_ht", True):
           return False

       if not dataset.HasColumn("reiteration", "base azi", True):
           return False

       if not dataset.HasColumn("reiteration", "top_azi", True):
           return False

       if not dataset.HasColumn("reiteration", "basal_diameter", True):
           return False

       if not dataset.HasColumn("reiteration", "base dist", True):
           return False

       if not dataset.HasColumn("reiteration", "top_dist", True):
           return False

       #Stem Table
       if not dataset.HasColumn("stem", "site_id", True):
           return False

       if not dataset.HasColumn("stem", "stem_id", True):
           return False

       if not dataset.HasColumn("stem", "dbh", True):
           return False

       if not dataset.HasColumn("stem", "height", True):
           return False

       if not dataset.HasColumn("stem", "branch_data?", True):
           return False

       if not dataset.HasColumn("stem", "live_status", True):
           return False

       #Branch Table
       if not dataset.HasColumn("branch", "site_id", True):
           return False

       if not dataset.HasColumn("branch", "branch_id", True):
           return False

       if not dataset.HasColumn("branch", "stem_id", True):
           return False

       if not dataset.HasColumn("branch", "live_status", True):
           return False

       if not dataset.HasColumn("branch", "height", True):
           return False

       if not dataset.HasColumn("branch", "diameter", True):
           return False

       if not dataset.HasColumn("branch", "azimuth", True):
           return False

       if not dataset.HasColumn("branch", "extent", True):
           return False

       if not dataset.HasColumn("branch", "slope", True):
           return False

       if not dataset.HasColumn("branch", "epicormic", True):
           return False

       if not dataset.HasColumn("branch", "Percent_cover", True):
           return False

       #All There?
       return True

   def GetUnits(self, dataset):
       "Returns a list of stems that can be visualized given the available data."

       table    = dataset.GetTable("stem")
       units    = [ ]

       while 1:
           stem = table.NextRowD()

           if stem == None:
               break

           if stem['live_status'] == "L" and stem['branch_data?'] == 1:
               try:
                   units.append([(stem['site_id'],
int(stem['stem_id'])), str(stem['site_id']) + " " +
                       str(stem['stem_id'])])
               except ValueError:
                   pass

       return units

   def GetOptions(self, dataset):
       "Returns a list of options to be provided to the end user at run time."
       options = [
           CheckboxOption('grid', "Show Grid", True),
           CheckboxOption('rulers', "Show Rulers", False),
           CheckboxOption('info', "Show Tree Information", True),
           CheckboxOption('cover', "Scale Foliage by Recorded Percent Cover", True),
           CheckboxOption('reiteration', "Show Reiterations", True),
           CheckboxOption('dead', "Show Dead Branches", True),
           CheckboxOption('back', "Show Back Brancher", False),
           CheckboxOption('front', "Show Front Branches", False),
           CheckboxOption('hide_center', "Hide Center Line", False),
           SpacerOption(),
           ColorOption('bg_color', "Background Color", (255,255,255)),
           ColorOption('label_color', "Label Background Color", (0,0,0)),
           SpacerOption(),
           SliderOption('recursion', "Detail Level", 0, 6, 2),
           SliderOption('branchscale', "Exagerate Branch Sizes", 1, 200, 1),
           RadioButtonsOption('shading', "Shading Method", ["None",
"Opacity", "Intensity"]),
           ColorOption('trunk_ink_color', "Trunk Ink Color", (0,0,0)),
           ColorOption('branch_ink_color', "Branch Ink Color", (0, 0, 0)),
           ColorOption('foliage_ink_color', "Foliage Ink Color", (0, 0, 0)),
           ColorOption('reiteration_ink_color', "Reiteration Ink Color", (0,0,0))
       ]

       return options

   def GetUnitName(self, dataset, options, id, view=0):
       "Returns a filename for the given unit id and view with current dataset and options."

       return "Tree %d (%s)" % (id[1], id[0])

   def GetInverseColor(self, color):
       "Internal method. Get an inverse of an RGB color"

       return (abs(color[0]-255), abs(color[1]-255), abs(color[2]-255))

   def GetBranchQuadrant(self, camera, branch):
       "sort out where the branch is in relation to the viewer."
       if branch > 180:
           branch = branch - 360

       if branch >= -45 and branch < 45:
           return "back"
       if branch >= 45 and branch < 160:
           return "right"
       if branch >= -160 and branch < -45:
           return "left"
       else:
           return "front"

   def GetForeShortening(self, camera, branch):
       "returns a number between 0 and 1, indicating how much the extent needs to be foreshortened."

  #     print("foreshortening = " + str(math.sin(branch * math.pi / 180)))
       return abs(math.sin(branch * math.pi / 180))


   def DrawLeftABABranch(self, image, count, start, end, options,
startd = 1, endd = 1):
       "recursively draws a right-leaning context free branch"
       if count > options['recursion']:
           return
           #stop recursing!


       #image.Draw(OI_Line(start, end, .01, (0, 0, 0)))
       dx = end[0] - start[0]
       dy = end[1] - start[1]
       thirdpt = (start[0] + dx/3, start[1] + dy/3)
       twothirdpt = (start[0] + 2 * dx/3, start[1] +  2 * dy/3)
       thirdlength = math.sqrt(dx**2  + dy **2) / 3
       oldangle = math.atan2(dy, dx)

       #left branch
       langle = oldangle + kBranchAngle
       lendx = thirdpt[0] + thirdlength * (math.sin(langle))
       lendy = thirdpt[1] + thirdlength + (math.cos(langle))

       #right branch
       rangle = oldangle - kBranchAngle
       rendx = twothirdpt[0] - thirdlength * (math.sin (rangle))
       rendy = twothirdpt[1] - thirdlength + (math.cos(rangle))

       #recurse!
       self.DrawLeftABABranch(image, count+1, start, thirdpt, options)
       self.DrawLeftABABranch(image, count+1, thirdpt, twothirdpt, options)
       self.DrawLeftABABranch(image, count+1, twothirdpt, end, options)
       self.DrawLeftABABranch(image, count+1, thirdpt, (rendx, rendy), options)
       self.DrawLeftABABranch(image, count+1, twothirdpt, (lendx,
lendy), options)

       opac = 255
       if options['shading'] == 1:
           print("shading opacity!")
           opac = 255 * (options['recursion'] + 1 - count)/float(options['recursion'] + 2)
       print("count = " + str(count) + " opac = " + str(opac))

       image.Draw(OI_Line((start[0], start[1] - startd/2), (end[0],
end[1] - endd/2), .01, (0, 0, 0, opac)))
       image.Draw(OI_Line((start[0], start[1] + startd/2), (end[0],
end[1] + endd/2), .01, (0, 0, 0, opac)))




   def DrawRightABABranch(self, image, count, start, end, options,
startd = 1, endd = 1):
       "recursively draws a right-leaning context free branch"
       if count > options['recursion']:
           return

       #image.Draw(OI_Line(start, end, .01, (0, 0, 0)))
       dx = end[0] - start[0]
       dy = end[1] - start[1]
       thirdpt = (start[0] + dx/3, start[1] + dy/3)
       twothirdpt = (start[0] + 2 * dx/3, start[1] +  2 * dy/3)
       thirdlength = math.sqrt(dx**2  + dy **2) / 3
       oldangle = math.atan2(dy, dx)
       dd = startd - endd

       #right branch
       rangle = oldangle + kBranchAngle
       rendx = thirdpt[0] + thirdlength * (math.sin(rangle))
       rendy = thirdpt[1] + thirdlength + (math.cos(rangle))

       #left branch
       langle = oldangle - kBranchAngle
       lendx = twothirdpt[0] - thirdlength * (math.sin (langle))
       lendy = twothirdpt[1] - thirdlength + (math.cos(langle))

       #recurse!
       self.DrawRightABABranch(image, count+1, start, thirdpt,
options, startd, startd - dd/3)
       self.DrawRightABABranch(image, count+1, thirdpt, twothirdpt,
options, startd - dd/3, startd - 2*dd/3)
       self.DrawRightABABranch(image, count+1, twothirdpt, end,
options, startd - 2*dd/3, endd)
       self.DrawRightABABranch(image, count+1, thirdpt, (rendx,
rendy), options, (startd - dd/3)/2, 0)
       self.DrawRightABABranch(image, count+1, twothirdpt, (lendx,
lendy), options, (startd - 2*dd/3), 0)

       opac = 255
       if options['shading'] == 1:
           print("shading opacity!")
           opac = 255 * (options['recursion'] + 1 - count)/float(options['recursion'] + 2)
       print("count = " + str(count) + " opac = " + str(opac))

       image.Draw(OI_Line((start[0], start[1] - startd/2), (end[0],
end[1] - endd/2), .01, (0, 0, 0, opac)))
       image.Draw(OI_Line((start[0], start[1] + startd/2), (end[0],
end[1] + endd/2), .01, (0, 0, 0, opac)))




   def GetImage(self, dataset, options, id, view=0):
       "Returns an OmniImage instance representing the requested stem and view number."

       stem = self.GetStem(dataset, id[0], id[1])

       if not stem:
           return OmniImage(10,10)

       offset = [0,0]

       if options['rulers']:
           offset[0] += 50

       if options['info']:
           offset[1] += 30

       image = OmniImage((kBaseImageSize[0]+offset[0],
kBaseImageSize[1]+offset[1]), options['bg_color'])

       #Draw the grid
       if options['grid']:
           #Horizontal Lines
           for i in range(0, int(kBaseImageSize[1]/kCellSize)+1):
               y = kBaseImageSize[1] - (i*kCellSize)
               image.Draw(OI_Line((0 + offset[0], y),
(kBaseImageSize[0] + offset[0], y), 1,
                   self.GetInverseColor(options['bg_color'])))

           #Vertical Lines
           vLinesRange = range(0, int(kBaseImageSize[0]/kCellSize)+1)
           for i in vLinesRange:
               if not options['hide_center'] or len(vLinesRange)%2 == 0 or i != int(len(vLinesRange)/2):
                   x = offset[0] + (i*kCellSize)
                   image.Draw(OI_Line((x, 0), (x, kBaseImageSize[1]),
1, self.GetInverseColor(options['bg_color'])))

       #Draw the ruler
       if options['rulers']:
           for i in range(1, int(kBaseImageSize[1]/kCellSize)+1):
               y            = kBaseImageSize[1] - (i*kCellSize)
               label        = str(i * kCellSize/kPixelsToMeter) + "m"
               labelSize    = OI_String((0,0), label, 18).GetStringSize()
               corner        = (47 - labelSize[0], y - labelSize[1])

               image.Draw(OI_String(corner, label, 18, self.GetInverseColor(options['bg_color'])))

       #Draw tree information
       if options['info']:
           #Box
           start    = (offset[0], kBaseImageSize[1] - 30 + offset[1])
           end        = (kBaseImageSize[0] + offset[0], kBaseImageSize[1] + offset[1])

           image.Draw(OI_Rectangle(start, end, options['label_color'], line=1))

           #Label
           label        = "Tree %s (%s)" % (stem.id, stem.site)
           labelSize    = OI_String((0,0), label, 18).GetStringSize()
           corner        = (int(kBaseImageSize[0]/2 - labelSize[0]/2 + offset[0]), int(kBaseImageSize[1] - 15 -
                           labelSize[1]/2 + offset[1]))

           image.Draw(OI_String(corner, label, 18, self.GetInverseColor(options['label_color'])))

       arc            = 180
       rightCenter    = 180
       leftCenter    = 0
       rightSide    = (90,270)
       midPoint    = kBaseImageSize[0]/2 + offset[0]

       #Draw the branches behind the stem, if we are using them.
       if(options['back']):
           for branch in stem.branches:
               quad = self.GetBranchQuadrant(0, branch.azimuth)
               if(quad == "back"):
                   if branch.azimuth < 0:
                       multiplier = -1
                   else:
                       multiplier = 1
            
                   trunkedge = stem.GetRadiusAtHeight(branch.height) + 1

                   coverDist = (100 - float(branch.percentCover) )/100

                   if(not options['cover']):
                       coverDist = 0

                   slopeGain = math.atan(branch.slope * math.pi / 180.0) * branch.extent
                   extent = branch.extent * multiplier

                   extent = self.GetForeShortening(0, branch.azimuth) * extent

                   startpt = (midPoint + multiplier*trunkedge, kBaseImageSize[1] - branch.height*kPixelsToMeter)
                   endpt = (midPoint + extent*kPixelsToMeter,
                        kBaseImageSize[1] - (branch.height + branch.diameter/2 + slopeGain) * kPixelsToMeter)

                   coverpt = (startpt[0] + (endpt[0] -startpt[0]) * coverDist,
                          startpt[1] + (endpt[1] -startpt[1]) * coverDist)

                   if(not branch.alive and options['dead']):
                        image.Draw(OI_Line((startpt[0], startpt[1] + kPixelsToMeter*branch.diameter/2),
                                      endpt, .5, options['branch_ink_color']))
                        image.Draw(OI_Line((startpt[0], startpt[1] - kPixelsToMeter*branch.diameter/2),
                                      endpt, .5, options['branch_ink_color']))


                   elif(branch.alive and multiplier > 0):
                       self.DrawRightABABranch(image, 0, coverpt, endpt, options, coverDist*options['branchscale']*kPixelsToMeter*branch.diameter/float(100), 0)

                       image.Draw(OI_Line((startpt[0],startpt[1] + kPixelsToMeter*branch.diameter/2),
                                  (coverpt[0],coverpt[1] + kPixelsToMeter * branch.diameter/2*coverDist),
                                   .5, options['branch_ink_color']))
                       image.Draw(OI_Line((startpt[0],startpt[1] - kPixelsToMeter*branch.diameter/2),
                                  (coverpt[0],coverpt[1] - kPixelsToMeter * branch.diameter/2*coverDist),
                                   .5, options['branch_ink_color']))
                   else:
                       image.Draw(OI_Line((startpt[0],startpt[1] + kPixelsToMeter*branch.diameter/2),
                                  (coverpt[0],coverpt[1] + kPixelsToMeter * branch.diameter/2*coverDist),
                                   .5, options['branch_ink_color']))
                       image.Draw(OI_Line((startpt[0], startpt[1],  kPixelsToMeter*branch.diameter/2),
                                  (coverpt[0],coverpt[1] - kPixelsToMeter * branch.diameter/2*coverDist),
                                   .5, options['branch_ink_color']))


                       self.DrawLeftABABranch(image, 0, coverpt, endpt, options, coverDist*options['branchscale']*kPixelsToMeter*branch.diameter/float(100),0)


                    
            



       #Draw the stem
       baseRadius    = stem.diameter / 100 / 2 * kPixelsToMeter
       treeHeight    = stem.height * kPixelsToMeter

       #trunkPoly = [(midPoint - baseRadius, kBaseImageSize[1]),(midPoint + baseRadius, kBaseImageSize[1]),
       #                (midPoint + 1, kBaseImageSize[1] -treeHeight), (midPoint - 1, kBaseImageSize[1] - treeHeight)]
       #image.Draw(OI_Polygon(trunkPoly, (182, 61, 54)))
       image.Draw(OI_Polygon([(midPoint - baseRadius, kBaseImageSize[1]),
                              (midPoint + baseRadius, kBaseImageSize[1]),
                              (midPoint + 1, kBaseImageSize[1] - treeHeight)],
                              options['bg_color']))
       
       image.Draw(OI_Line((midPoint - baseRadius, kBaseImageSize[1]),
                          (midPoint + 1, kBaseImageSize[1] - treeHeight), .5,
                          options['trunk_ink_color']))
       image.Draw(OI_Line((midPoint + baseRadius, kBaseImageSize[1]),
                          (midPoint + 1, kBaseImageSize[1] - treeHeight), .5,
                          options['trunk_ink_color']))



       #Draw Branches
       for branch in stem.branches:


           quad = self.GetBranchQuadrant(0, branch.azimuth)
               

           if quad == "back" or quad == "front":
               continue

           coverDist = (100 - float(branch.percentCover) )/100
           if(not options['cover']):
               coverDist = 0


           slopeGain = math.atan(branch.slope * math.pi / 180.0) * branch.extent
           if quad == "right":
               extent = branch.extent * -1
           else:
               extent = branch.extent

           extent = self.GetForeShortening(0, branch.azimuth) * extent



           trunkedge = stem.GetRadiusAtHeight(branch.height) + 1
           #print("trunkedge = " + str(trunkedge) + " getrad = " + str(stem.GetRadiusAtHeight(branch.height)))

               

           if(quad == "right" and branch.comes_from == "M"):
               startpt = (midPoint - trunkedge, kBaseImageSize[1] - branch.height*kPixelsToMeter)
               endpt = (midPoint + extent*kPixelsToMeter,
                        kBaseImageSize[1] - (branch.height + branch.diameter/2 + slopeGain) * kPixelsToMeter)

               coverpt = (startpt[0] + (endpt[0] -startpt[0]) * coverDist,
                          startpt[1] + (endpt[1] -startpt[1]) * coverDist)

               if(not branch.alive and options['dead']):
                   image.Draw(OI_Line((startpt[0], startpt[1] + kPixelsToMeter*branch.diameter/2),
                                      endpt, .5, options['branch_ink_color']))
                   image.Draw(OI_Line((startpt[0], startpt[1] - kPixelsToMeter*branch.diameter/2),
                                      endpt, .5, options['branch_ink_color']))


               elif(branch.alive):
                   self.DrawRightABABranch(image, 0, coverpt, endpt, options, coverDist*options['branchscale']*kPixelsToMeter*branch.diameter/float(100), 0)

                   image.Draw(OI_Line((startpt[0],startpt[1] + kPixelsToMeter*branch.diameter/2),
                                  (coverpt[0],coverpt[1] + kPixelsToMeter * branch.diameter/2*coverDist),
                                   .5, options['branch_ink_color']))
                   image.Draw(OI_Line((startpt[0],startpt[1] - kPixelsToMeter*branch.diameter/2),
                                  (coverpt[0],coverpt[1] - kPixelsToMeter * branch.diameter/2*coverDist),
                                   .5, options['branch_ink_color']))




               #self.DrawRightABABranch(image, 0, (midPoint - trunkedge, kBaseImageSize[1] - branch.height*kPixelsToMeter),
               #                        (midPoint + extent*kPixelsToMeter, kBaseImageSize[1] - (branch.height +branch.diameter/2 +
               #                                slopeGain) *kPixelsToMeter), options,options['branchscale']*kPixelsToMeter*branch.diameter/float(100), 0)



           if(quad == "left" and branch.comes_from == "M"):

               startpt = (midPoint + trunkedge, kBaseImageSize[1] - branch.height*kPixelsToMeter)
               endpt = (midPoint + extent*kPixelsToMeter, kBaseImageSize[1] - (branch.height + branch.diameter/2 +
                       slopeGain) * kPixelsToMeter)

               coverpt = (startpt[0] + (endpt[0] -startpt[0]) * coverDist,
                          startpt[1] + (endpt[1] -startpt[1]) * coverDist)

               if(not branch.alive and options['dead']):
                   image.Draw(OI_Line((startpt[0], startpt[1] + kPixelsToMeter*branch.diameter/2),
                                      endpt, .5, options['branch_ink_color']))
                   image.Draw(OI_Line((startpt[0], startpt[1] - kPixelsToMeter*branch.diameter/2),
                                      endpt, .5, options['branch_ink_color']))


               elif(branch.alive):    
                   image.Draw(OI_Line((startpt[0],startpt[1] + kPixelsToMeter*branch.diameter/2),
                                  (coverpt[0],coverpt[1] + kPixelsToMeter * branch.diameter/2*coverDist),
                                   .5, options['branch_ink_color']))
                   image.Draw(OI_Line((startpt[0], startpt[1],  kPixelsToMeter*branch.diameter/2),
                                  (coverpt[0],coverpt[1] - kPixelsToMeter * branch.diameter/2*coverDist),
                                   .5, options['branch_ink_color']))


                   self.DrawLeftABABranch(image, 0, coverpt, endpt, options, coverDist*options['branchscale']*kPixelsToMeter*branch.diameter/float(100),0)



       #Draw Reiterations
       if options['reiteration']:
           for reit in stem.reiterations:
               print("drawing a reiteration")
               quad = self.GetBranchQuadrant(0, reit.azimuth)

               trunkedge = stem.GetRadiusAtHeight(reit.base) + 1
               trunkedge  = trunkedge * kPixelsToMeter


               if quad == "front" or quad == "back":
                   continue



               elif quad == "right":
                   distance = reit.distance * -1
                   diameter = reit.diameter * -1
                   trunkedge = trunkedge * -1
                   base = reit.distance * -1
                   modifier = -1
               else:
                   distance = reit.distance
                   diameter = reit.diameter
                   base = reit.distance
                   modifier= 1


               trunkedge = trunkedge/100
               bottomtrunk = (midPoint + trunkedge, kBaseImageSize[1] - reit.base*kPixelsToMeter)
               toptrunk = (midPoint + trunkedge, kBaseImageSize[1] - reit.base*kPixelsToMeter - reit.diameter*kPixelsToMeter/100)
               basept = (midPoint + base*kPixelsToMeter,kBaseImageSize[1] - reit.base*kPixelsToMeter)
               elbowpt = (basept[0] + modifier*kPixelsToMeter / 100 * reit.diameter, basept[1] - reit.diameter/100 * kPixelsToMeter)
               toppt = (basept[0],kBaseImageSize[1] - reit.top*kPixelsToMeter)

               print("top = " + str(toppt))
               print("base = " + str(base))

               image.Draw(OI_Line(bottomtrunk, basept, 1, options['reiteration_ink_color']))
               image.Draw(OI_Line(basept, toppt, 1, options['reiteration_ink_color']))
               image.Draw(OI_Line(toppt, elbowpt, 1, options['reiteration_ink_color']))
               image.Draw(OI_Line(elbowpt, toptrunk, 1, options['reiteration_ink_color']))

               for branch in stem.branches:
                   if branch.comes_from == reit.name:
                       print("FOUND A REIT BRANCH!")

               #image.Draw(OI_Line(bottomtrunk, basept, 5, (255, 0, 0)))
               #image.Draw(OI_Line(bottomtrunk, toppt, 5, (0, 255, 0)))


       #        points = [(midPoint + distance*kPixelsToMeter,kBaseImageSize[1] - reit.base*kPixelsToMeter),
       #                    (midPoint + distance*kPixelsToMeter +diameter*kPixelsToMeter, kBaseImageSize[1] -
       #                    reit.base*kPixelsToMeter), (midPoint +distance*kPixelsToMeter + diameter/2*kPixelsToMeter,
       #                    kBaseImageSize[1] - reit.top*kPixelsToMeter)]

       #        image.Draw(OI_Polygon(points, options['reiteration_color']))
       #        image.Draw(OI_Rectangle((midPoint, kBaseImageSize[1] - reit.base*kPixelsToMeter), (midPoint +
       #            distance*kPixelsToMeter +diameter*kPixelsToMeter, kBaseImageSize[1] -
       #            (reit.base-reit.diameter)*kPixelsToMeter), options['reiteration_color']))

       #finally, draw front branches if we're using them.
       if(options['front']):
           for branch in stem.branches:
               quad = self.GetBranchQuadrant(0, branch.azimuth)
               if(quad == "front"):
                   if branch.azimuth > 180:
                       multiplier = -1
                   else:
                       multiplier = 1
            
                   trunkedge = stem.GetRadiusAtHeight(branch.height) + 1

                   coverDist = (100 - float(branch.percentCover) )/100

                   if(not options['cover']):
                       coverDist = 0

                   slopeGain = math.atan(branch.slope * math.pi / 180.0) * branch.extent
                   extent = branch.extent * multiplier

                   extent = self.GetForeShortening(0, branch.azimuth) * extent

                   startpt = (midPoint + multiplier*trunkedge, kBaseImageSize[1] - branch.height*kPixelsToMeter)
                   endpt = (midPoint + extent*kPixelsToMeter,
                        kBaseImageSize[1] - (branch.height + branch.diameter/2 + slopeGain) * kPixelsToMeter)

                   coverpt = (startpt[0] + (endpt[0] -startpt[0]) * coverDist,
                          startpt[1] + (endpt[1] -startpt[1]) * coverDist)

                   if(not branch.alive and options['dead']):
                        image.Draw(OI_Line((startpt[0], startpt[1] + kPixelsToMeter*branch.diameter/2),
                                      endpt, .5, options['branch_ink_color']))
                        image.Draw(OI_Line((startpt[0], startpt[1] - kPixelsToMeter*branch.diameter/2),
                                      endpt, .5, options['branch_ink_color']))


                   elif(branch.alive and multiplier > 0):
                       self.DrawRightABABranch(image, 0, coverpt, endpt, options, coverDist*options['branchscale']*kPixelsToMeter*branch.diameter/float(100), 0)

                       image.Draw(OI_Line((startpt[0],startpt[1] + kPixelsToMeter*branch.diameter/2),
                                  (coverpt[0],coverpt[1] + kPixelsToMeter * branch.diameter/2*coverDist),
                                   .5, options['branch_ink_color']))
                       image.Draw(OI_Line((startpt[0],startpt[1] - kPixelsToMeter*branch.diameter/2),
                                  (coverpt[0],coverpt[1] - kPixelsToMeter * branch.diameter/2*coverDist),
                                   .5, options['branch_ink_color']))
                   else:
                       image.Draw(OI_Line((startpt[0],startpt[1] + kPixelsToMeter*branch.diameter/2),
                                  (coverpt[0],coverpt[1] + kPixelsToMeter * branch.diameter/2*coverDist),
                                   .5, options['branch_ink_color']))
                       image.Draw(OI_Line((startpt[0], startpt[1],  kPixelsToMeter*branch.diameter/2),
                                  (coverpt[0],coverpt[1] - kPixelsToMeter * branch.diameter/2*coverDist),
                                   .5, options['branch_ink_color']))


                       self.DrawLeftABABranch(image, 0, coverpt, endpt, options, coverDist*options['branchscale']*kPixelsToMeter*branch.diameter/float(100),0)
       
       return image

class Stem:
   "Tree struct."

   def __init__(self):
       "Constructor."

       self.id                = 0
       self.site            = ""
       self.height            = 0.0
       self.diameter        = 0.0
       self.alive            = False
       self.hasBranchData    = False
       self.branches        = [ ]
       self.reiterations    = [ ]


   def GetRadiusAtHeight(self, height):
       "this function returns the radius of the tree at a specified height"
       #need to take taper into account.
       return (self.diameter * (self.height / float(height)) ) / 200


class Branch:
   "Branch struct."

   def __init__(self):
       "Constructor."

       self.id                = 0
       self.stem            = 0
       self.site            = ""
       self.height            = 0.0
       self.extent            = 0.0
       self.azimuth        = 0
       self.diameter        = 0.0
       self.alive            = False
       self.percentCover    = 0
       self.slope            = 0
       self.epicormic        = False
       self.comes_from      = "M"

class Reiteration:
   "Reiteration struct."

   def __init__(self):
       "Constructor."

       self.id            = 0
       self.stem        = 0
       self.site        = ""
       self.azimuth    = 0
       self.base        = 0.0
       self.top        = 0.0
       self.diameter    = 0.0
       self.distance    = 0.0
       self.name        = ""