# CanopyView 2D Plugins - neighbor.py
# Description: Neighbor (example) plugin.
# Contributors: Nik Molnar
# Created: 2009-06-12
# (c)2009 The Evergreen State College

from Plugins import *
from OmniImage import *

def init():
	"Return an instance of the plugin class."
	
	return NeighborPlugin()

class NeighborPlugin:
	"Example plugin. Shows a tree's surrounding neighbors."
	
	def Info(self):
		"Plugin information."
		
		info = {
			'name'			: "Neighbors",
			'description'	: "Example plugin. Shows a tree's surrounding neighbors.",
			'unit_label'	: "Stem",
			'views'			: 1
		}
		
		return info
		
	def RequiredType(self):
		"Returns 'sql', 'flat', or None depending on required dataset type."
		
		return 'sql'
		
	def RequiredTables(self):
		"Returns a list of required tables, or None if expecting a table-less dataset."
		
		return ["stem"]
		
	def RequiredColumns(self):
		"Returns a list of required columns in the form [(table, column), ...]"
		
		return [("stem", "site_id"), ("stem", "stem_id"), ("stem", "locationxyz_x"), 
				("stem", "locationxyz_y"), ("stem", "dbh")]
		
	def ValidData(self, dataset):
		'''Returns True if the provided dataset contains the data necessary to generate this visualization. 
		   False otherwise.'''
		
		#Since we use a SQL query, this dataset must be of type 'sql'
		if dataset.GetType() != 'sql':
			return False
			
		#Check for required columns
		if not dataset.HasColumn("stem", "site_id", True):
			return False
			
		if not dataset.HasColumn("stem", "stem_id", True):
			return False
			
		if not dataset.HasColumn("stem", "locationxyz_x", True):
			return False
			
		if not dataset.HasColumn("stem", "locationxyz_y", True):
			return False
			
		if not dataset.HasColumn("stem", "dbh", True):
			return False
			
		#If everything is there, return True
		return True
		
	def GetUnits(self, dataset):
		"Returns a list of ids and labels for units from the dataset that can be visualized given the available data."
		
		units = [ ]
		
		#Get a result set for the stem table
		stems = dataset.GetTable("stem")
		
		#For each stem
		while 1:
			stem = stems.NextRowD()
			
			if stem == None:
				break
				
			#Here, we define ID as a tuple containing the stem_id and the site_id
			id = (stem['site_id'], stem['stem_id'])
			
			#Add the id to the list of units with a label
			units.append([id, "Tree #%s (%s)" % (stem['stem_id'], stem['site_id'])])
			
		return units
		
	def GetOptions(self, dataset):
		"Returns a list of options to be provided to the end user at run time."
		
		options = [ ]
		
		#Extent determines how much area (in meters) around the specified stem is shown
		options.append(NumberOption('extent', "Extent", 2.5))
		
		#Spacer -- separate extent option visually from the color options
		options.append(SpacerOption())
		
		#Image background color -- default is white
		options.append(ColorOption('background_color', "Background", (255,255,255)))
		
		#Selected stem color -- default is green
		options.append(ColorOption('selected_color', "Selected Stem", (0,255,0)))
		
		#Other stem color -- default is brown
		options.append(ColorOption('stem_color', "Other Stems", (153,102,51)))
		
		return options
		
	def GetUnitName(self, dataset, options, id, view=0):
		"Returns a filename for the given unit id and view with current dataset and options."
		
		#Return a name containing stem and site ids.
		return "Stem %s (%s)" % (id[1], id[0])
		
	def GetImage(self, dataset, options, id, view=0):
		'''Returns an OmniImage instance representing the given view for the unit specified by id given the 
		   provided data and options.'''
		   
		#Create a new 1000x1000 image with the user-specified background color
		image = OmniImage(size=(1000,1000), background=options['background_color'])
		
		#Get the specified stem
		result = dataset.ExecQuery("SELECT %s, %s, %s FROM %s WHERE %s = \"%s\" AND %s = %s" % (
			dataset.MapColumn("locationxyz_x"), dataset.MapColumn("locationxyz_y"), dataset.MapColumn("dbh"), 
			dataset.MapTable("stem"), dataset.MapColumn("site_id"), id[0], dataset.MapColumn("stem_id"), id[1]))
		
		#Set stem information or return a blank image if value errors.
		try:
			stem	= result.NextRow()
			
			if not stem:
				return image
			
			locationX	= float(stem[0])
			locationY	= float(stem[1])
			dbh			= float(stem[2]) / 100 #DBH is given in cm, so divide by 100 to get m
			
		except ValueError:
			return image
			
		#Determine ratio of meters to pixels
		scale = 1000 / (options['extent'] * 2)
		
		#Determine where (0,0) for this square is in real-word coordinates (m)
		offsetX	= locationX - options['extent']
		offsetY	= locationY - options['extent']
			
		#Determine pixel locations for x
		xPixStart	= scale * (locationX - offsetX - dbh/2)
		xPixEnd		= scale * (locationX - offsetX + dbh/2)
		
		#Determine pixel locations for y
		yPixStart	= scale * (locationY - offsetY - dbh/2)
		yPixEnd		= scale * (locationY - offsetY + dbh/2)

		#Define the stem, but wait to draw it so that it ends up on top
		stemEllipse = OI_Ellipse((xPixStart, yPixStart), (xPixEnd, yPixEnd), options['selected_color'])
		
		#Query near-by stems
		result = dataset.ExecQuery("SELECT %s, %s, %s FROM %s WHERE %s = \"%s\" \
					AND %s > %f AND %s < %f AND %s > %f AND %s < %f" % (
					dataset.MapColumn("locationxyz_x"), dataset.MapColumn("locationxyz_y"),
					dataset.MapColumn("dbh"), dataset.MapTable("stem"), dataset.MapColumn("site_id"),
					id[0], dataset.MapColumn("locationxyz_x"), locationX - options['extent'], 
					dataset.MapColumn("locationxyz_x"), locationX + options['extent'],
					dataset.MapColumn("locationxyz_y"), locationY - options['extent'], 
					dataset.MapColumn("locationxyz_y"), locationY + options['extent']))
					
		#For each result
		while 1:
			stem = result.NextRow()
			
			if not stem:
				break
				
			try:
				locationX	= float(stem[0])
				locationY	= float(stem[1])
				dbh			= float(stem[2]) / 100 #DBH is given in cm, so divide by 100 to get m
				
			except ValueError:
				continue
				
			#Determine pixel locations for x
			xPixStart	= scale * (locationX - offsetX - dbh/2)
			xPixEnd		= scale * (locationX - offsetX + dbh/2)
			
			#Determine pixel locations for y
			yPixStart	= scale * (locationY - offsetY - dbh/2)
			yPixEnd		= scale * (locationY - offsetY + dbh/2)

			#Draw the stem
			image.Draw(OI_Ellipse((xPixStart, yPixStart), (xPixEnd, yPixEnd), options['stem_color']))
			
		#Draw the selected stem
		image.Draw(stemEllipse)
			
		#Return the image
		return image
		